// created by Josh Heyde
// last updated 27/01/09

#include "C_FlockingEntity.h"
#include "C_Rectangle.h"
#include "C_Visualisation.h"

#include <assert.h>
#include <cmath>

// defines null rather than including the entire cstdlib header file
#define NULL 0

// constructor
C_FlockingEntity::C_FlockingEntity(const int &ID, const int &cF, const float &pX, const float &pY, const int &s, const int &rR, const int &mV)
: C_WorldEntity(ID, cF, pX, pY, s), repulsionRange(rR), MAX_VELOCITY(mV)
{
	direction = new C_Vector(1, 1);
}

// destructor
C_FlockingEntity::~C_FlockingEntity()
{
	delete direction;
}

// mutator functions
void C_FlockingEntity::setRepulsionRange(const int &newRepulsionRange)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newRepulsionRange <= 0);

	repulsionRange = newRepulsionRange;
}

void C_FlockingEntity::setDirection(const int &newX, const int &newY)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newX < 0);
	assert(newY < 0);

	direction->x = newX;
	direction->y = newY;
}

void C_FlockingEntity::setDirection(const C_Vector* newDirection)
{
	// error handling statements
	// asserting certain statements for this function
	assert(newDirection != NULL);

	*direction = *newDirection;
}

// other member functions
void C_FlockingEntity::update()
{
	C_WorldEntity::update();

	// prevents the entity going off screen
	if(nextPosition->x > VIS.getScreenWidth())
	{
		direction->x = -MAX_VELOCITY;
	}

	if(nextPosition->x < 0)
	{
		direction->x = MAX_VELOCITY;
	}

	if(nextPosition->y > VIS.getScreenHeight())
	{
		direction->y = -MAX_VELOCITY;
	}

	if(nextPosition->y < 0)
	{
		direction->y = MAX_VELOCITY;
	}

	// moves the entity
	*nextPosition += (*direction * speed);
}

// draws the sprite
void C_FlockingEntity::render(const float &currentTime)
{
	// error handling statements
	// asserting certain statements for this function
	assert(currentTime >= 0);
	
	C_Vector inter(currentPosition->x, currentPosition->y);

	// only interpolates if the next position is different to the current position
	if((currentPosition->x != nextPosition->x) || (currentPosition->y != nextPosition->y))
	{
		inter.x += currentTime * (nextPosition->x - currentPosition->x);
		inter.y += currentTime * (nextPosition->y - currentPosition->y);
	}

	VIS.drawSprite(graphicID, (int) inter.x, (int) inter.y, currentFrame);
}

// determines the direction the entity will move in
void C_FlockingEntity::determineDirection(std::vector <C_WorldEntity*> v_entityVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(!v_entityVector.empty());

	// creates variables to calculate the direction to move
	C_Vector temp(0, 0);

	// models the aligment rule
	temp += alignment(v_entityVector);
	// models the cohesion rule
	temp += cohesion(v_entityVector);
	// models the separation rule
	temp += avoidance(v_entityVector);

	// adjusts the direction of the entity to follow the
	// the suggested directions from each of
	// the implemented behaviours
	*direction = temp;

	// ensures the entity doesn't move too fast
	controlVelocity();
}

// controls the velocity of the entity
void C_FlockingEntity::controlVelocity()
{
	if((fabs(direction->x) > MAX_VELOCITY) || (fabs(direction->y) > MAX_VELOCITY))
	{
		direction->Normalise();
		*direction *= MAX_VELOCITY;
	}
}

////////////////////////
// behavioural functions
////////////////////////

// seeks to avoid entities that are too close to this entity
C_Vector C_FlockingEntity::avoidance(std::vector<C_WorldEntity*> v_entityVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(!v_entityVector.empty());

	// creates a temporary vector to work out the direction to separate
	C_Vector temp(0, 0);

	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		if(v_entityVector[i] != this)
		{
			// adds up all the vectors between this entity and
			// the surrounding entities
			if(((*currentPosition - *v_entityVector[i]->getCurrentPosition()).Length()) < repulsionRange)
			{
				temp += (*currentPosition - *v_entityVector[i]->getCurrentPosition());
			}
		}
	}

	return temp;
}

// seeks to align the direction of this entity with those surrounding it
C_Vector C_FlockingEntity::alignment(std::vector<C_WorldEntity*> v_entityVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(!v_entityVector.empty());

	// creates a temporary vector to work out the direction to separate
	C_Vector temp(0, 0);
	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		if(v_entityVector[i] != this)
		{
			// adds up all the vectors between this entity and
			// the surrounding entities
			temp += *v_entityVector[i]->getDirection();
		}
	}

	// calculates the average direction of nearby entities
	// and follows that
	temp /= (v_entityVector.size() -1);

	return temp;
}

// seeks to aim towards the average position of the surrounding entities
C_Vector C_FlockingEntity::cohesion(std::vector<C_WorldEntity*> v_entityVector)
{
	// error handling statements
	// asserting certain statements for this function
	assert(!v_entityVector.empty());

	// creates a temporary vector to work out the direction
	C_Vector temp(0, 0);
	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		if(v_entityVector[i] != this)
		{
			// adds up all the current positions of the surrounding entities
			temp += *v_entityVector[i]->getCurrentPosition();
		}
	}

	// sets the entity to aim for the average position of the nearby entities
	temp /= (v_entityVector.size() - 1);
	temp -= *currentPosition;

	return temp;
}