// created by Josh Heyde
// last updated 25/01/09

#include "C_Sprite.h"
#include "C_Rectangle.h"
#include <HAPI_lib.h>
#include <assert.h>

// constructor
C_Sprite::C_Sprite(const int &numOfF, const int &fw, const int &fh, const bool &t)
: numOfFrames(numOfF), frameWidth(fw), frameHeight(fh), transparency(t), texture(0)
{
}

// default constructor
C_Sprite::C_Sprite(): width(0), height(0), texture(0), numOfFrames(0), frameWidth(0), frameHeight(0)
{
}

// destructor
C_Sprite::~C_Sprite()
{
	delete[] texture;
}

bool C_Sprite::initialise(const std::string &fileName)
{	
	// error handling statements
	// asserting certain statements for this function
	assert(fileName.length() != 0);
	
	if(!HAPI->LoadTexture(fileName, &texture, &width, &height))
	{
		 HAPI->UserMessage("\nLoad error with texture:\n" + fileName, "Texture Load Error");
		 delete[] texture;
		 return false;
	}

	return true;
}

void C_Sprite::render(BYTE *dest, const C_Rectangle &R_dest, int posX, int posY, int &currentFrame)
{
	// error handling statements
	// asserting certain statements for this function
	assert(dest != 0);
	assert(texture != 0);

	// creating a rectangle to hold the source
	C_Rectangle R_source(0, width, 0, height);

	// creating a temporary rectangle to clip
	C_Rectangle R_temp(0, frameWidth, 0, frameHeight);
	
	// exits the function if the texture is completely off the screen
	// (no need to process)
	if(R_temp.allExternal(R_dest, posX, posY))
		return;

	// step 1: setting the current frame to render
	if(currentFrame > numOfFrames - 1)
		currentFrame = 0;

	// step 2: converting this rectangle to destination space
	R_temp.left += posX;
	R_temp.right += posX;
	R_temp.top += posY;
	R_temp.bottom += posY;
	
	// blits as normal unless the texture is partially off screen
	// if the texture is partially off screen, it is clipped
	if(!R_temp.allInternal(R_dest, posX, posY))
	{
		// step 3: clipping against the destination
		R_temp.clip(R_dest);
	}
		
	// step 4: converting back to source space
	R_temp.left -= posX;
	R_temp.right -= posX;
	R_temp.top -= posY;
	R_temp.bottom -= posY;

	// step 5: clipping any negative values
	if(R_temp.left > frameWidth)
		R_temp.left = frameWidth;

	if(R_temp.right < 0)
		R_temp.right = 0;

	if(R_temp.top > frameHeight)
		R_temp.top = frameHeight;

	if(R_temp.bottom < 0)
		R_temp.bottom = 0;

	if(posX < 0)
		posX = 0;

	if(posY < 0)
		posY = 0;

	// step 6: advancing the frame
	R_temp.left += frameWidth * currentFrame;
	R_temp.right += frameWidth * currentFrame;
		
	// setting up the pointers to the position of the texture on the screen
	// as well as the top left corner of texture
	BYTE *destPoint = dest + (posX + posY * R_dest.getWidth()) * 4;
	BYTE *sourcePoint = texture + (R_temp.left + R_temp.top * R_source.getWidth()) * 4;

	// uses the correct method of blitting depending on transparency settings
	if(transparency)
		blitAlpha(destPoint, sourcePoint, R_dest, R_source, R_temp);
	else
		blitNoAlpha(destPoint, sourcePoint, R_dest, R_source, R_temp);
}

void C_Sprite::blitNoAlpha(BYTE *destPoint, BYTE *sourcePoint, const C_Rectangle &R_dest, const C_Rectangle &R_source, const C_Rectangle &R_temp)
{
	// error handling statements
	// asserting certain statements for this function
	assert(destPoint != 0);
	assert(sourcePoint != 0);
	
	// loading in a texture by copying it over line by line
	// does not handle transparency
	for(int y = 0; y < R_temp.getHeight(); y++)
	{
		memcpy(destPoint, sourcePoint, R_temp.getWidth() * 4);

		// incrementing the pointers
		destPoint += R_dest.getWidth() * 4;
		sourcePoint += R_source.getWidth() * 4;
	}
}

void C_Sprite::blitAlpha(BYTE *destPoint, BYTE *sourcePoint, const C_Rectangle &R_dest, const C_Rectangle &R_source, const C_Rectangle &R_temp)
{
	// error handling statements
	// asserting certain statements for this function
	assert(destPoint != 0);
	assert(sourcePoint != 0);

	// a variable to ensure the texture is drawn to the correct location
	// as this method is done pixel by pixel
	int endOfLineDestIncrement = (R_dest.getWidth() - R_temp.getWidth()) * 4;

	// variable to ensure that texture pointer jumps to the next line at the correct
	// time when the texture is partially on screen
	int endOfLineSourceIncrement = (R_source.getWidth() - R_temp.getWidth()) * 4;

	for(int y = 0; y < R_temp.getHeight(); y++)
	{
		for(int x = 0; x < R_temp.getWidth(); x++)
		{
			BYTE blue = sourcePoint[0];
			BYTE green = sourcePoint[1];
			BYTE red = sourcePoint[2];
			BYTE alpha = sourcePoint[3];

			// optimisations for transparency display
			// if aplpha is 0, no need to display,
			// if alpha is 255, then destination will be set to source
			if(alpha != 0)
				if(alpha == 255)
				{
					destPoint[0] = blue;
					destPoint[1] = green;
					destPoint[2] = red;
				}
				else
				{
					destPoint[0] += (alpha * (blue - destPoint[0])) >> 8;
					destPoint[1] += (alpha * (green - destPoint[1])) >> 8;
					destPoint[2] += (alpha * (red - destPoint[2])) >> 8;
				}

			// incrementing the pointers
			destPoint += 4;
			sourcePoint += 4;
		}
		// incrementing the pointers to the next line
		destPoint += endOfLineDestIncrement;
		sourcePoint += endOfLineSourceIncrement;
	}
}

// mutator functions

void C_Sprite::setWidth(const int &newWidth)
{
	width = newWidth;
}

void C_Sprite::setHeight(const int &newHeight)
{
	height = newHeight;
}

void C_Sprite::setNumOfFrames(const int &newNumOfFrames)
{
	numOfFrames = newNumOfFrames;
}

void C_Sprite::setFrameWidth(const int &newFrameWidth)
{
	frameWidth = newFrameWidth;
}

void C_Sprite::setFrameHeight(const int &newFrameHeight)
{
	frameHeight = newFrameHeight;
}

void C_Sprite::setTransparency(const bool &newTransparency)
{
	transparency = newTransparency;
}