#include "C_Visualisation.h"
#include "C_Sprite.h"
#include "C_Rectangle.h"
#include <HAPI_lib.h>
#include <assert.h>

// initialisation of the singleton
C_Visualisation* C_Visualisation::vis = NULL;

// creation of a singleton
void C_Visualisation::createVisual(const int &sw, const int &sh)
{
	if(!vis)
		vis = new C_Visualisation(sw, sh);
}

// deletion of a singleton
void C_Visualisation::deleteVisual()
{
	// error handling statements
	// asserting certain statements for this function
	assert(vis != NULL);

	delete vis;
}


// constructor
C_Visualisation::C_Visualisation(const int &sw, const int &sh) : screenWidth(sw), screenHeight(sh), screen(0)
{
}

C_Visualisation::~C_Visualisation()
{
	for (std::vector<C_Sprite*>::const_iterator iter = v_spriteVector.begin(); iter < v_spriteVector.end(); iter++)
		delete *iter;
}

bool C_Visualisation::initialise()
{
	// error handling statements
	// asserting certain statements for this function
	assert(screenWidth > 0);
	assert(screenHeight > 0);
	
	// ensuring the initialisation occurred correctly
     if (!HAPI->Initialise(&screenWidth,&screenHeight))
	 {
		 delete screen;
         return false;
	 }

	 // retrieves the correct position of the pointer to the screen buffer
	 screen = HAPI->GetScreenPointer();
	 HAPI->SetShowFPS(true);

	 return true;
}

void C_Visualisation::clearToColour(const BYTE &r, const BYTE &g, const BYTE &b, BYTE a) const
{
	// error handling statements
	// asserting certain statements for this function
	assert(screen != 0);
	assert(screenWidth > 0);
	assert(screenHeight > 0);
	assert(r <= 255 && r >= 0);
	assert(g <= 255 && g >= 0);
	assert(b <= 255 && b >= 0);
	assert(a <= 255 && a >= 0);

	HAPI_TColour col(r, g, b, a);
	BYTE *tempScreen = screen;

	// a small optimisation increasing the frame rate
	// if the colour to be cleared to is grayscale
	if(r == g && g == b && r == b && a == 255)
		memset(tempScreen, r, screenWidth * screenHeight * 4);
	else
	{
		for(int y = 0; y < screenHeight; y++)
		{
			for(int x = 0; x < screenWidth; x++)
			{
				int offset = (x + y * screenWidth) * 4;
				BYTE *currentLocation = tempScreen + offset;
				memcpy(currentLocation, &col, 4);
			}
		}
	}
}

bool C_Visualisation::createSprite(const std::string &filename, int &id, const int &numOfFrames, const int &frameWidth, const int &frameHeight, const bool &transparency)
{
	// error handling statements
	// asserting certain statements for this function
	assert(filename.length() != 0);

	C_Sprite *newSprite = new C_Sprite(numOfFrames, frameWidth, frameHeight, transparency);
	if(!newSprite->initialise(filename))
	{
		delete newSprite;
		return false;
	}

	// storing the sprite
	v_spriteVector.push_back(newSprite);
	id = v_spriteVector.size() - 1;
	return true;
}

void C_Visualisation::drawSprite(const int &id, const int &posX, const int &posY, int &currentFrame) const
{
	// creating the rectangle objects for the textures
	// for clipping
	C_Rectangle R_dest(0, screenWidth, 0, screenHeight);

	v_spriteVector[id]->render(screen, R_dest, posX, posY, currentFrame);
}

void C_Visualisation::displayText(const int &posX, const int &posY, std::string text, const BYTE &r, const BYTE &g, const BYTE &b, BYTE a) const
{
	// error handling statements
	// asserting certain statements for this function
	assert(r <= 255 && r >= 0);
	assert(g <= 255 && g >= 0);
	assert(b <= 255 && b >= 0);
	assert(a <= 255 && a >= 0);

	HAPI_TColour col(r, g, b, a);
	HAPI->RenderText(posX, posY, col, text);
}

// mutator functions
void C_Visualisation::setScreenWidth(const int &newScreenWidth)
{
	screenWidth = newScreenWidth;
}

void C_Visualisation::setScreenHeight(const int &newScreenHeight)
{
	screenHeight = newScreenHeight;
}