// created by Josh Heyde
// last updated 25/01/09

#include "C_World.h"
#include "C_WorldEntity.h"
#include "C_FlockingEntity.h"
#include "C_Visualisation.h"
#include "C_UI.h"
#include <HAPI_lib.h>
#include <assert.h>

#include <sstream>

using std::string;

// initialisation of the singleton
C_World* C_World::world = NULL;

// creation of a singleton
void C_World::createWorld()
{
	if(!world)
		world = new C_World();
}

// deletion of a singleton
void C_World::deleteWorld()
{
	// error handling statements
	// asserting certain statements for this function
	assert(world != NULL);

	delete world;
}

// constructor
C_World::C_World() : currentTime(0), previousTime(0), DELAY(50)
{
	// setting up the screen size
	 const int screenWidth = 800;
     const int screenHeight = 600;

	// initialises the visualisation singleton
	VIS.createVisual(screenWidth, screenHeight);

	// creates the UI singleton
	INPUT.createInput();
}

// destructor
C_World::~C_World()
{
	// deletes all created world entities
	for (std::vector<C_WorldEntity*>::const_iterator iter = v_entityVector.begin(); iter < v_entityVector.end(); iter++)
		delete *iter;

	// deletes the visualisation singleton
	VIS.deleteVisual();

	// deletes the UI singleton
	INPUT.deleteInput();
}

// other member functions
bool C_World::initialise()
{
	// checking that the program initialises correctly
	if(!VIS.initialise())
	{
		// deletes all created world entities
		for (std::vector<C_WorldEntity*>::const_iterator iter = v_entityVector.begin(); iter < v_entityVector.end(); iter++)
			delete *iter;
		VIS.deleteVisual();
		INPUT.deleteInput();

		return false;
	}

	return true;
}

void C_World::loadData()
{
	int ID = 0;

	const int NUM_OF_ENTITIES = 20;

	if(!VIS.createSprite("enemyBullet.png", ID, 0, 16, 16))
		return;

	// creating the flockers
	for(int i = 0; i < NUM_OF_ENTITIES; i++)
		v_entityVector.push_back(new C_FlockingEntity(ID, 0, rand() % (VIS.getScreenWidth() - 16), rand() % (VIS.getScreenHeight() - 16),
		5, 30, 5));
}

template <class T>
std::string toString(const T & t)
{
    std::ostringstream oss; // create a stream
    oss << t;              // insert value to stream
    return oss.str();      // return as a string
}

// calls the UI component to get input
void C_World::getInput()
{
	// closes HAPI if esc is pressed
	if(INPUT.isEscPressed())
	{
		HAPI->Close();
	}
}

// renders the sprites of the world entities
void C_World::draw(const float &currentTick)
{
	// refreshes the screen by clearing to black
	VIS.clearToColour(0, 0, 0);

	// rendering each entity
	for(size_t i = 0; i < v_entityVector.size(); i++)
	{
		if(v_entityVector[i]->getActive())
		{
			v_entityVector[i]->render(currentTick);
		}
	}

	// displays useful info on screen
	VIS.displayText(0, 13, "Press Esc to Exit", 255, 255, 255);	
}

void C_World::update()
{
	currentTime = HAPI->GetTime();

	// gets input from the user
	getInput();

	// only updates the world entities after a set period of time
	if(currentTime > (previousTime + DELAY))
	{
		// updates every entity
		for(size_t i = 0; i < v_entityVector.size(); i++)
		{
			// only updates the entity if it is active
			if(v_entityVector[i]->getActive())
			{
				// evaluates the direction of every entity
				v_entityVector[i]->determineDirection(v_entityVector);
				
				v_entityVector[i]->update();
			}
		}

		previousTime = currentTime;
	}

	draw((float)(currentTime - previousTime) / DELAY);	
}

